@@ -10,6 +10,7 @@ from ...sim_type import (SimTypeLongLong, SimTypeInt, SimTypeShort, SimTypeChar,
     SimTypeBottom, SimTypeArray, SimTypeFunction)
 from ...sim_variable import SimVariable, SimTemporaryVariable, SimStackVariable, SimRegisterVariable, SimMemoryVariable
 from ...utils.constants import is_alignment_mask
+from ...utils.library import get_cpp_function_name
 from ...errors import UnsupportedNodeTypeError
 from .. import Analysis, register_analysis
 from .region_identifier import MultiNode
@@ -35,7 +36,7 @@ class PositionMappingElement:
         return self.start <= offset < self.start + self.length
 
     def __repr__(self):
-        return "<%d-%d: %s>" % (self.start, self.start + self.length, self.obj.c_repr())
+        return "<%d-%d: %s>" % (self.start, self.start + self.length, self.obj)
 
 
 class PositionMapping:
@@ -178,7 +179,11 @@ class CFunction(CConstruct):  # pylint:disable=abstract-method
         yield self.functy.returnty.c_repr(), None
         yield " ", None
         # function name
-        yield self.demangled_name or self.name, None
+        if self.demangled_name:
+            normalized_name = get_cpp_function_name(self.demangled_name, specialized=False, qualified=False)
+        else:
+            normalized_name = self.name
+        yield normalized_name, None
         # argument list
         yield "(", None
         for i, (arg_type, arg) in enumerate(zip(self.functy.args, self.arg_list)):
@@ -206,6 +211,31 @@ class CStatement(CConstruct):  # pylint:disable=abstract-method
     __slots__ = ()
 
 
+class CExpression:
+    """
+    Base class for C expressions.
+    """
+
+    __slots__ = ('_type', )
+
+    def __init__(self):
+        self._type = None
+
+    @property
+    def type(self):
+        raise NotImplementedError("Class %s does not implement type()." % type(self))
+
+    def set_type(self, v):
+        self._type = v
+
+    @staticmethod
+    def _try_c_repr_chunks(expr):
+        if hasattr(expr, 'c_repr_chunks'):
+            yield from expr.c_repr_chunks()
+        else:
+            yield str(expr), None
+
+
 class CStatements(CStatement):
     """
     Represents a sequence of statements in C.
@@ -488,16 +518,16 @@ class CAssignment(CStatement):
         yield ";\n", None
 
 
-class CFunctionCall(CStatement):
+class CFunctionCall(CStatement, CExpression):
     """
     func(arg0, arg1)
 
     :ivar Function callee_func:  The function getting called.
     """
 
-    __slots__ = ('callee_target', 'callee_func', 'args', 'returning', 'ret_expr', )
+    __slots__ = ('callee_target', 'callee_func', 'args', 'returning', 'ret_expr', 'tags', 'is_expr', )
 
-    def __init__(self, callee_target, callee_func, args, returning=True, ret_expr=None):
+    def __init__(self, callee_target, callee_func, args, returning=True, ret_expr=None, tags=None, is_expr: bool=False):
         super().__init__()
 
         self.callee_target = callee_target
@@ -505,6 +535,16 @@ class CFunctionCall(CStatement):
         self.args = args if args is not None else [ ]
         self.returning = returning
         self.ret_expr = ret_expr
+        self.tags = tags
+        self.is_expr = is_expr
+
+    @property
+    def type(self):
+        if self.is_expr:
+            # TODO: Return the proper type of the ret_expr's
+            return SimTypeInt(signed=False)
+        else:
+            raise RuntimeError("CFunctionCall.type should not be accessed if the function call is used as a statement.")
 
     def c_repr_chunks(self, indent=0):
 
@@ -512,12 +552,15 @@ class CFunctionCall(CStatement):
 
         yield indent_str, None
 
-        if self.ret_expr is not None:
+        if not self.is_expr and self.ret_expr is not None:
             yield from CExpression._try_c_repr_chunks(self.ret_expr)
             yield " = ", None
 
         if self.callee_func is not None:
-            func_name = self.callee_func.demangled_name or self.callee_func.name
+            if self.callee_func.demangled_name:
+                func_name = get_cpp_function_name(self.callee_func.demangled_name, specialized=False, qualified=True)
+            else:
+                func_name = self.callee_func.name
             yield func_name, self
         else:
             yield from CExpression._try_c_repr_chunks(self.callee_target)
@@ -529,12 +572,13 @@ class CFunctionCall(CStatement):
                 yield ", ", None
             yield from CExpression._try_c_repr_chunks(arg)
 
-        yield ");", None
+        yield ")", None
 
-        if not self.returning:
-            yield " /* do not return */", None
-
-        yield "\n",  None
+        if not self.is_expr:
+            yield ";", None
+            if not self.returning:
+                yield " /* do not return */", None
+            yield "\n",  None
 
 
 class CReturn(CStatement):
@@ -600,31 +644,6 @@ class CUnsupportedStatement(CStatement):
         yield "\n", None
 
 
-class CExpression:
-    """
-    Base class for C expressions.
-    """
-
-    __slots__ = ('_type', )
-
-    def __init__(self):
-        self._type = None
-
-    @property
-    def type(self):
-        raise NotImplementedError("Class %s does not implement type()." % type(self))
-
-    def set_type(self, v):
-        self._type = v
-
-    @staticmethod
-    def _try_c_repr_chunks(expr):
-        if hasattr(expr, 'c_repr_chunks'):
-            yield from expr.c_repr_chunks()
-        else:
-            yield str(expr), None
-
-
 class CStructField(CExpression):
 
     __slots__ = ('struct_type', 'offset', 'field', )
@@ -724,14 +743,26 @@ class CVariable(CExpression):
                                     yield "->", None
                                     yield from c_field.c_repr_chunks()
                                     return
+                                else:
+                                    # accessing beyond known offset - indicates a bug in type inference
+                                    l.warning("Accessing non-existent offset %d in struct %s. This indicates a bug in "
+                                              "the type inference engine.", self.offset, self.variable.type.pts_to)
 
                         elif isinstance(self.variable.type.pts_to, SimTypeArray):
-                            # it's pointing to an array!
-                            yield from self.variable.c_repr_chunks()
-                            yield "[", None
-                            yield str(self.offset), self.offset
-                            yield "]", None
-                            return
+                            if isinstance(self.offset, int):
+                                # it's pointing to an array! take the corresponding element
+                                yield from self.variable.c_repr_chunks()
+                                yield "[", None
+                                yield str(self.offset), self.offset
+                                yield "]", None
+                                return
+
+                        # other cases
+                        yield from self.variable.c_repr_chunks()
+                        yield "[", None
+                        yield from CExpression._try_c_repr_chunks(self.offset)
+                        yield "]", None
+                        return
 
                 # default output
                 yield "*(", None
@@ -813,9 +844,9 @@ class CBinaryOp(CExpression):
     Binary operations.
     """
 
-    __slots__ = ('op', 'lhs', 'rhs', 'variable', )
+    __slots__ = ('op', 'lhs', 'rhs', 'variable', 'tags', )
 
-    def __init__(self, op, lhs, rhs, variable):
+    def __init__(self, op, lhs, rhs, variable, tags: Optional[dict]=None):
 
         super().__init__()
 
@@ -823,6 +854,7 @@ class CBinaryOp(CExpression):
         self.lhs = lhs
         self.rhs = rhs
         self.variable = variable
+        self.tags = tags
 
     @property
     def type(self):
@@ -932,32 +964,32 @@ class CBinaryOp(CExpression):
 
     def _c_repr_chunks_cmple(self):
         yield from self._try_c_repr_chunks(self.lhs)
-        yield " <= ", None
+        yield " <= ", self
         yield from self._try_c_repr_chunks(self.rhs)
 
     def _c_repr_chunks_cmplt(self):
         yield from self._try_c_repr_chunks(self.lhs)
-        yield " < ", None
+        yield " < ", self
         yield from self._try_c_repr_chunks(self.rhs)
 
     def _c_repr_chunks_cmpgt(self):
         yield from self._try_c_repr_chunks(self.lhs)
-        yield " > ", None
+        yield " > ", self
         yield from self._try_c_repr_chunks(self.rhs)
 
     def _c_repr_chunks_cmpge(self):
         yield from self._try_c_repr_chunks(self.lhs)
-        yield " >= ", None
+        yield " >= ", self
         yield from self._try_c_repr_chunks(self.rhs)
 
     def _c_repr_chunks_cmpeq(self):
         yield from self._try_c_repr_chunks(self.lhs)
-        yield " == ", None
+        yield " == ", self
         yield from self._try_c_repr_chunks(self.rhs)
 
     def _c_repr_chunks_cmpne(self):
         yield from self._try_c_repr_chunks(self.lhs)
-        yield " != ", None
+        yield " != ", self
         yield from self._try_c_repr_chunks(self.rhs)
 
 
@@ -1233,22 +1265,25 @@ class StructuredCodeGenerator(Analysis):
     # Handlers
     #
 
-    def _handle(self, node):
+    def _handle(self, node, is_expr: bool=True):
         handler = self._handlers.get(node.__class__, None)
         if handler is not None:
+            # special case for Call
+            if isinstance(node, Stmt.Call):
+                return handler(node, is_expr=is_expr)
             return handler(node)
         raise UnsupportedNodeTypeError("Node type %s is not supported yet." % type(node))
 
     def _handle_Code(self, node):
 
-        return self._handle(node.node)
+        return self._handle(node.node, is_expr=False)
 
     def _handle_Sequence(self, seq):
 
         lines = [ ]
 
         for node in seq.nodes:
-            lines.append(self._handle(node))
+            lines.append(self._handle(node, is_expr=False))
 
         if not lines:
             return CStatements([])
@@ -1259,11 +1294,11 @@ class StructuredCodeGenerator(Analysis):
 
         if loop_node.sort == 'while':
             return CWhileLoop(None if loop_node.condition is None else self._handle(loop_node.condition),
-                              self._handle(loop_node.sequence_node)
+                              self._handle(loop_node.sequence_node, is_expr=False)
                               )
         elif loop_node.sort == 'do-while':
             return CDoWhileLoop(self._handle(loop_node.condition),
-                                self._handle(loop_node.sequence_node)
+                                self._handle(loop_node.sequence_node, is_expr=False)
                                 )
 
         else:
@@ -1272,8 +1307,10 @@ class StructuredCodeGenerator(Analysis):
     def _handle_Condition(self, condition_node):
 
         code = CIfElse(self._handle(condition_node.condition),
-                       true_node=self._handle(condition_node.true_node) if condition_node.true_node else None,
-                       false_node=self._handle(condition_node.false_node) if condition_node.false_node else None,
+                       true_node=self._handle(condition_node.true_node, is_expr=False)
+                                 if condition_node.true_node else None,
+                       false_node=self._handle(condition_node.false_node, is_expr=False)
+                                  if condition_node.false_node else None,
                        )
         return code
 
@@ -1290,7 +1327,7 @@ class StructuredCodeGenerator(Analysis):
         lines = [ ]
 
         for n in node.nodes:
-            r = self._handle(n)
+            r = self._handle(n, is_expr=False)
             lines.append(r)
 
         return CStatements(lines) if len(lines) > 1 else lines[0]
@@ -1303,8 +1340,8 @@ class StructuredCodeGenerator(Analysis):
         """
 
         switch_expr = self._handle(node.switch_expr)
-        cases = [ (idx, self._handle(case)) for idx, case in node.cases.items() ]
-        default = self._handle(node.default_node) if node.default_node is not None else None
+        cases = [ (idx, self._handle(case, is_expr=False)) for idx, case in node.cases.items() ]
+        default = self._handle(node.default_node, is_expr=False) if node.default_node is not None else None
         switch_case = CSwitchCase(switch_expr, cases, default=default)
         return switch_case
 
@@ -1323,7 +1360,7 @@ class StructuredCodeGenerator(Analysis):
         cstmts = [ ]
         for stmt in node.statements:
             try:
-                cstmt = self._handle(stmt)
+                cstmt = self._handle(stmt, is_expr=False)
             except UnsupportedNodeTypeError:
                 l.warning("Unsupported AIL statement or expression %s.", type(stmt), exc_info=True)
                 cstmt = CUnsupportedStatement(stmt)
@@ -1346,12 +1383,24 @@ class StructuredCodeGenerator(Analysis):
             # special handling
             base, offset = None, None
             if isinstance(cvariable, CBinaryOp) and cvariable.op == 'Add':
+                # variable and a const
                 if isinstance(cvariable.lhs, CConstant) and isinstance(cvariable.rhs, CVariable):
                     offset = cvariable.lhs.value
                     base = cvariable.rhs
                 elif isinstance(cvariable.rhs, CConstant) and isinstance(cvariable.lhs, CVariable):
                     offset = cvariable.rhs.value
                     base = cvariable.lhs
+                # variable and a typecast
+                elif isinstance(cvariable.lhs, CVariable) and isinstance(cvariable.rhs, CTypeCast):
+                    offset = cvariable.rhs
+                    base = cvariable.lhs
+                elif isinstance(cvariable.rhs, CVariable) and isinstance(cvariable.lhs, CTypeCast):
+                    offset = cvariable.lhs
+                    base = cvariable.rhs
+                elif isinstance(cvariable.lhs, CVariable) and isinstance(cvariable.rhs, CVariable):
+                    # GUESS: we need some guessing here
+                    base = cvariable.lhs
+                    offset = cvariable.rhs
                 else:
                     base = None
                     offset = None
@@ -1375,7 +1424,7 @@ class StructuredCodeGenerator(Analysis):
 
         return CAssignment(cdst, csrc)
 
-    def _handle_Stmt_Call(self, stmt):
+    def _handle_Stmt_Call(self, stmt, is_expr: bool=False):
 
         try:
             # Try to handle it as a normal function call
@@ -1430,6 +1479,8 @@ class StructuredCodeGenerator(Analysis):
         return CFunctionCall(target, target_func, args,
                              returning=target_func.returning if target_func is not None else True,
                              ret_expr=ret_expr,
+                             tags=stmt.tags,
+                             is_expr=is_expr,
                              )
 
     def _handle_Stmt_Jump(self, stmt):
@@ -1491,6 +1542,7 @@ class StructuredCodeGenerator(Analysis):
 
         return CBinaryOp(expr.op, lhs, rhs,
                          variable=self._handle(expr.variable) if expr.variable is not None else None,
+                         tags=expr.tags,
                          )
 
     def _handle_Expr_Convert(self, expr):
