@@ -9,7 +9,7 @@ from ...calling_conventions import DEFAULT_CC, SimRegArg, SimStackArg
 from ...knowledge_plugins.key_definitions.atoms import Register, Tmp, MemoryLocation
 from ...knowledge_plugins.key_definitions.constants import OP_BEFORE, OP_AFTER
 from ...knowledge_plugins.key_definitions.dataset import DataSet
-from ...knowledge_plugins.key_definitions.undefined import Undefined, undefined
+from ...knowledge_plugins.key_definitions.undefined import Undefined, UNDEFINED
 from ...knowledge_plugins.key_definitions.live_definitions import Definition
 from .external_codeloc import ExternalCodeLocation
 from .rd_state import ReachingDefinitionsState
@@ -21,17 +21,19 @@ class SimEngineRDAIL(
     SimEngineLightAILMixin,
     SimEngineLight,
 ):  # pylint:disable=abstract-method
-    def __init__(self, project, current_local_call_depth, maximum_local_call_depth, function_handler=None):
+    def __init__(self, project, call_stack, maximum_local_call_depth, function_handler=None):
         super(SimEngineRDAIL, self).__init__()
         self.project = project
-        self._current_local_call_depth = current_local_call_depth
+        self._call_stack = call_stack
         self._maximum_local_call_depth = maximum_local_call_depth
         self._function_handler = function_handler
         self._visited_blocks = None
+        self._dep_graph = None
 
         self.state: ReachingDefinitionsState
 
     def process(self, state, *args, **kwargs):
+        self._dep_graph = kwargs.pop('dep_graph', None)
         self._visited_blocks = kwargs.pop('visited_blocks', None)
 
         # we are using a completely different state. Therefore, we directly call our _process() method before
@@ -45,7 +47,7 @@ class SimEngineRDAIL(
         except SimEngineError as e:
             if kwargs.pop('fail_fast', False) is True:
                 raise e
-        return self.state, self._visited_blocks
+        return self.state, self._visited_blocks, self._dep_graph
 
     #
     # Private methods
@@ -86,7 +88,7 @@ class SimEngineRDAIL(
         dst = stmt.dst
 
         if src is None:
-            src = DataSet(undefined, dst.bits)
+            src = DataSet(UNDEFINED, dst.bits)
 
         if isinstance(dst, ailment.Tmp):
             self.state.kill_and_add_definition(Tmp(dst.tmp_idx, dst.size), self._codeloc(), src)
@@ -117,7 +119,7 @@ class SimEngineRDAIL(
                 l.info('Memory address undefined, ins_addr = %#x.', self.ins_addr)
                 continue
 
-            if any(type(d) is Undefined for d in data):
+            if isinstance(data, DataSet) and any(type(d) is Undefined for d in data):
                 l.info('Data to write at address %s undefined, ins_addr = %#x.',
                        hex(a) if type(a) is int else a, self.ins_addr
                        )
@@ -147,7 +149,10 @@ class SimEngineRDAIL(
             self.state.kill_definitions(Register(*self.arch.registers['cc_dep2']), self._codeloc())
             self.state.kill_definitions(Register(*self.arch.registers['cc_ndep']), self._codeloc())
 
-    def _ail_handle_Call(self, stmt):
+    def _ail_handle_Call(self, stmt: ailment.Stmt.Call):
+        self._handle_Call_base(stmt, is_expr=False)
+
+    def _handle_Call_base(self, stmt: ailment.Stmt.Call, is_expr: bool=False):
         target = self._expr(stmt.target)  # pylint:disable=unused-variable
 
         ip = Register(self.arch.ip_offset, self.arch.bytes)
@@ -181,7 +186,6 @@ class SimEngineRDAIL(
             cc = self.project.factory.cc()
 
         killed_vars = [ Register(*self.arch.registers[reg_name]) for reg_name in cc.CALLER_SAVED_REGS ]
-        return_reg_offset, return_reg_size = self.arch.registers[cc.RETURN_VAL.reg_name]
 
         # Add uses
         if used_exprs is None:
@@ -189,8 +193,23 @@ class SimEngineRDAIL(
         for expr in used_exprs:
             self._expr(expr)
 
-        # Return value is redefined here, so it is not a dummy value
-        self.state.kill_definitions(Register(return_reg_offset, return_reg_size), self._codeloc(), dummy=False)
+        # Add definition
+        return_reg_offset = None
+        if not is_expr:
+            if stmt.ret_expr is not None:
+                if isinstance(stmt.ret_expr, ailment.Expr.Register):
+                    return_reg_offset = stmt.ret_expr.reg_offset
+                    return_reg_size = stmt.ret_expr.size
+                    self.state.kill_and_add_definition(Register(return_reg_offset, return_reg_size), self._codeloc(),
+                                                       None)
+                else:
+                    l.warning("Unsupported ret_expr type %s. Please report to GitHub.", stmt.ret_expr.__class__)
+
+            else:
+                # Return value is redefined here, so it is not a dummy value
+                return_reg_offset, return_reg_size = self.arch.registers[cc.RETURN_VAL.reg_name]
+                self.state.kill_definitions(Register(return_reg_offset, return_reg_size), self._codeloc(), dummy=False)
+
         # Kill those ones that should be killed
         for var in killed_vars:
             if var.reg_offset == return_reg_offset:
@@ -254,6 +273,10 @@ class SimEngineRDAIL(
 
         return super(SimEngineRDAIL, self)._ail_handle_Tmp(expr)
 
+    def _ail_handle_CallExpr(self, expr: ailment.Stmt.Call):
+        self._handle_Call_base(expr, is_expr=True)
+        return DataSet(UNDEFINED, expr.bits)
+
     def _ail_handle_Register(self, expr):
 
         reg_offset = expr.reg_offset
@@ -328,18 +351,22 @@ class SimEngineRDAIL(
                 if current_defs:
                     for current_def in current_defs:
                         # self.state.add_use(current_def, codeloc)
-                        data.update(current_def.data)
+                        if isinstance(current_def.data, DataSet):
+                            data.update(current_def.data)
+                        else:
+                            # dropped
+                            l.warning("Dropping data of type %s since it is not a DataSet.", type(current_def.data))
                     if any(type(d) is Undefined for d in data):
                         l.info('Stack access at offset %#x undefined, ins_addr = %#x.', addr.offset, self.ins_addr)
                 else:
-                    data.add(undefined)
+                    data.add(UNDEFINED)
 
                 self.state.add_use(MemoryLocation(addr, size), self._codeloc())
             else:
                 l.debug('Memory address %r undefined or unsupported at pc %#x.', addr, self.ins_addr)
 
         if len(data) == 0:
-            data.add(undefined)
+            data.add(UNDEFINED)
 
         return DataSet(data, bits)
 
@@ -382,7 +409,7 @@ class SimEngineRDAIL(
                 r = DataSet(converted, expr.to_bits)
 
         if r is None:
-            r = DataSet(undefined, expr.to_bits)
+            r = DataSet(UNDEFINED, expr.to_bits)
 
         return r
 
@@ -390,12 +417,12 @@ class SimEngineRDAIL(
         cond = self._expr(expr.cond)
         iftrue = self._expr(expr.iftrue)
         iffalse = self._expr(expr.iffalse)
-        return ailment.Expr.ITE(expr.idx, cond, iffalse, iftrue)
+        return DataSet(ailment.Expr.ITE(expr.idx, cond, iffalse, iftrue), expr.bits)
 
     def _ail_handle_BinaryOp(self, expr):
         r = super()._ail_handle_BinaryOp(expr)
         if isinstance(r, ailment.Expr.BinaryOp):
-            return DataSet(undefined, r.bits)
+            return DataSet(UNDEFINED, r.bits)
         return r
 
     def _ail_handle_Cmp(self, expr):
@@ -426,14 +453,14 @@ class SimEngineRDAIL(
                        )
 
     def _ail_handle_DirtyExpression(self, expr):  # pylint:disable=no-self-use
-        return DataSet(undefined, expr.bits // 8)
+        return DataSet(UNDEFINED, expr.bits // 8)
 
     #
     # User defined high-level statement handlers
     #
 
     def _handle_function(self):
-        if self._current_local_call_depth > self._maximum_local_call_depth:
+        if len(self._call_stack) + 1 > self._maximum_local_call_depth:
             l.warning('The analysis reached its maximum recursion depth.')
             return None
 
@@ -473,11 +500,19 @@ class SimEngineRDAIL(
         elif is_internal is True:
             handler_name = 'handle_local_function'
             if hasattr(self._function_handler, handler_name):
-                is_updated, state = getattr(self._function_handler, handler_name)(self.state, ip_addr,
-                                                                                  self._current_local_call_depth + 1,
-                                                                                  self._maximum_local_call_depth)
+                is_updated, state, visited_blocks, dep_graph = getattr(self._function_handler, handler_name)(
+                    self.state,
+                    ip_addr,
+                    self._call_stack,
+                    self._maximum_local_call_depth,
+                    self._visited_blocks,
+                    self._dep_graph,
+                )
+
                 if is_updated is True:
                     self.state = state
+                    self._visited_blocks = visited_blocks
+                    self._dep_graph = dep_graph
             else:
                 l.warning('Please implement the local function handler with your own logic.')
         else:
