@@ -3,7 +3,7 @@ import logging
 import claripy
 import archinfo
 from archinfo import RegisterName
-from typing import Union, Optional, List, Dict
+from typing import Union, Optional, List, Dict, Type
 
 from .sim_type import SimType
 from .sim_type import SimTypeChar
@@ -512,7 +512,7 @@ class SimCC:
                 sizes = [ ]
                 for a in args:
                     if isinstance(a, SimType):
-                        if a.size is NotImplemented:
+                        if isinstance(a, SimTypeFixedSizeArray) or a.size is NotImplemented:
                             sizes.append(self.arch.bytes)
                         else:
                             sizes.append(a.size // 8)  # SimType.size is in bits
@@ -1061,6 +1061,7 @@ class SimCCMicrosoftAMD64(SimCC):
     FP_RETURN_VAL = SimRegArg('xmm0', 32)
     RETURN_ADDR = SimStackArg(0, 8)
     ARCH = archinfo.ArchAMD64
+    STACK_ALIGNMENT = 16
 
 class SimCCX86LinuxSyscall(SimCC):
     ARG_REGS = ['ebx', 'ecx', 'edx', 'esi', 'edi', 'ebp']
@@ -1170,7 +1171,7 @@ class SimCCAMD64WindowsSyscall(SimCC):
 class SimCCARM(SimCC):
     ARG_REGS = [ 'r0', 'r1', 'r2', 'r3' ]
     FP_ARG_REGS = []    # TODO: ???
-    CALLER_SAVED_REGS = [ 'r0', 'r1', 'r2', 'r3' ]
+    CALLER_SAVED_REGS = []
     RETURN_ADDR = SimRegArg('lr', 4)
     RETURN_VAL = SimRegArg('r0', 4)
     ARCH = archinfo.ArchARM
@@ -1394,7 +1395,7 @@ CC = {
 }
 
 
-DEFAULT_CC: Dict[str,type] = {
+DEFAULT_CC: Dict[str,Type[SimCC]] = {
     'AMD64': SimCCSystemVAMD64,
     'X86': SimCCCdecl,
     'ARMEL': SimCCARM,
