@@ -1,5 +1,5 @@
-
 # pylint:disable=no-self-use
+from typing import Tuple
 
 import logging
 
@@ -23,6 +23,7 @@ class SimEngineLight(SimEngine):
         self.state = None
         self.arch = None
         self.block = None
+        self._call_stack = None
 
         self.stmt_idx = None
         self.ins_addr = None
@@ -46,10 +47,20 @@ class SimEngineLight(SimEngine):
     # Helper methods
     #
 
+    @property
+    def _context(self) -> Tuple[int]:
+        if not self._call_stack:
+            return tuple()
+
+        # Convert to Tuple to make `context` hashable if not None
+        call_stack_addresses = tuple(self._call_stack)
+        return call_stack_addresses
+
     def _codeloc(self, block_only=False):
         return CodeLocation(self.block.addr,
                             None if block_only else self.stmt_idx,
                             ins_addr=None if block_only else self.ins_addr,
+                            context=self._context
                             )
 
 
@@ -92,6 +103,13 @@ class SimEngineLightVEXMixin:
 
             self._handle_Stmt(stmt)
 
+        self._process_block_end()
+
+    def _process_block_end(self):
+        # handle calls to another function
+        # Note that without global information, we cannot handle cases where we *jump* to another function (jumpkind ==
+        # "Ijk_Boring"). Users are supposed to overwrite this method, detect these cases with the help of global
+        # information (such as CFG or symbol addresses), and handle them accordingly.
         if self.block.vex.jumpkind == 'Ijk_Call':
             self.stmt_idx = DEFAULT_STATEMENT
             handler = '_handle_function'
@@ -207,6 +225,8 @@ class SimEngineLightVEXMixin:
         handler = None
         if expr.op.startswith('Iop_And'):
             handler = '_handle_And'
+        elif expr.op.startswith('Iop_Mod'):
+            handler = '_handle_Mod'
         elif expr.op.startswith('Iop_Or'):
             handler = '_handle_Or'
         elif expr.op.startswith('Iop_Add'):
@@ -236,11 +256,14 @@ class SimEngineLightVEXMixin:
             handler = '_handle_CmpLT'
         elif expr.op.startswith('Iop_CmpLE'):
             handler = '_handle_CmpLE'
+        elif expr.op.startswith('Iop_CmpGE'):
+            handler = '_handle_CmpGE'
+        elif expr.op.startswith('Iop_CmpGT'):
+            handler = '_handle_CmpGT'
         elif expr.op.startswith('Iop_CmpORD'):
             handler = '_handle_CmpORD'
         elif expr.op.startswith('Const'):
             handler = '_handle_Const'
-
         if handler is not None and hasattr(self, handler):
             return getattr(self, handler)(expr)
         else:
@@ -405,6 +428,25 @@ class SimEngineLightVEXMixin:
         except TypeError as e:
             self.l.warning(e)
             return None
+        except ZeroDivisionError as e:
+            self.l.warning(e)
+            return None
+
+    def _handle_Mod(self, expr):
+        arg0, arg1 = expr.args
+        expr_0 = self._expr(arg0)
+        if expr_0 is None:
+            return None
+        expr_1 = self._expr(arg1)
+        if expr_1 is None:
+            return None
+
+        try:
+            # TODO: Probably should take care of the sign
+            return expr_0 - (expr_0 // expr_1)*expr_1
+        except TypeError as e:
+            self.l.warning(e)
+            return None
 
     def _handle_Xor(self, expr):
         arg0, arg1 = expr.args
@@ -423,18 +465,24 @@ class SimEngineLightVEXMixin:
 
     def _handle_Shl(self, expr):
         arg0, arg1 = expr.args
+
         expr_0 = self._expr(arg0)
         if expr_0 is None:
             return None
         expr_1 = self._expr(arg1)
         if expr_1 is None:
             return None
-
+        #elif expr_1 < 0:
+        #    expr.args = [expr_0, -1*expr_1]
+         #   return self._handle_Shr(expr)
         try:
             if isinstance(expr_0, int) and isinstance(expr_1, int):
                 # self.tyenv is not used
                 mask = (1 << expr.result_size(self.tyenv)) - 1
-                return (expr_0 << expr_1) & mask
+                if expr_1 < 0:
+                    return (expr_0 >> -expr_1) & mask
+                else:
+                    return (expr_0 << expr_1) & mask
             else:
                 return expr_0 << expr_1
         except TypeError as e:
@@ -451,7 +499,10 @@ class SimEngineLightVEXMixin:
             return None
 
         try:
-            return expr_0 >> expr_1
+            if expr_1 < 0:
+                return expr_0 << -expr_1
+            else:
+                return expr_0 >> expr_1
         except TypeError as e:
             self.l.warning(e)
             return None
@@ -516,6 +567,21 @@ class SimEngineLightVEXMixin:
             self.l.warning(ex)
             return None
 
+    def _handle_CmpGE(self, expr):
+        arg0, arg1 = expr.args
+        expr_0 = self._expr(arg0)
+        if expr_0 is None:
+            return None
+        expr_1 = self._expr(arg1)
+        if expr_1 is None:
+            return None
+
+        try:
+            return expr_0 >= expr_1
+        except TypeError as ex:
+            self.l.warning(ex)
+            return None
+
     def _handle_CmpLT(self, expr):
         arg0, arg1 = expr.args
         expr_0 = self._expr(arg0)
@@ -531,6 +597,21 @@ class SimEngineLightVEXMixin:
             self.l.warning(ex)
             return None
 
+    def _handle_CmpGT(self, expr):
+        arg0, arg1 = expr.args
+        expr_0 = self._expr(arg0)
+        if expr_0 is None:
+            return None
+        expr_1 = self._expr(arg1)
+        if expr_1 is None:
+            return None
+
+        try:
+            return expr_0 > expr_1
+        except TypeError as ex:
+            self.l.warning(ex)
+            return None
+
     def _handle_MBE(self, expr):  # pylint:disable=unused-argument
         # Yeah.... no.
         return None
@@ -565,9 +646,24 @@ class SimEngineLightAILMixin:
 
     def _expr(self, expr):
 
-        handler = "_ail_handle_%s" % type(expr).__name__
+        expr_type_name = type(expr).__name__
+        if isinstance(expr, ailment.Stmt.Call):
+            # Call can be both an expression and a statement. Add a suffix to make sure we are working on the expression
+            # variant.
+            expr_type_name += "Expr"
+
+        h = None
+        handler = "_handle_%s" % expr_type_name
         if hasattr(self, handler):
-            return getattr(self, handler)(expr)
+            h = getattr(self, handler)
+
+        if h is None:
+            handler = "_ail_handle_%s" % expr_type_name
+            if hasattr(self, handler):
+                h = getattr(self, handler)
+
+        if h is not None:
+            return h(expr)
         self.l.warning('Unsupported expression type %s.', type(expr).__name__)
         return None
 
@@ -576,7 +672,7 @@ class SimEngineLightAILMixin:
     #
 
     def _codeloc(self):
-        return CodeLocation(self.block.addr, self.stmt_idx, ins_addr=self.ins_addr)
+        return CodeLocation(self.block.addr, self.stmt_idx, ins_addr=self.ins_addr, context=self._context)
 
     #
     # Statement handlers
@@ -620,6 +716,9 @@ class SimEngineLightAILMixin:
     def _ail_handle_Load(self, expr):
         raise NotImplementedError('Please implement the Load handler with your own logic.')
 
+    def _ail_handle_CallExpr(self, expr):
+        raise NotImplementedError('Please implement the CallExpr handler with your own logic.')
+
     def _ail_handle_UnaryOp(self, expr):
         handler_name = '_ail_handle_%s' % expr.op
         try:
