@@ -43,7 +43,7 @@ class SimEngineUnicorn(SuccessorsMixin):
 
         # if we have a concrete target we want the program to synchronize the segment
         # registers before, otherwise undefined behavior could happen.
-        if state.project.concrete_target:
+        if state.project.concrete_target and self.project.arch.name in ('x86', 'x86_64'):
             if not state.concrete.segment_registers_initialized:
                 l.debug("segment register must be synchronized with the concrete target before using unicorn engine")
                 return False
