@@ -4,7 +4,7 @@ import networkx
 import string
 import itertools
 from collections import defaultdict
-from typing import Union, Optional, Iterable
+from typing import Union, Optional, Iterable, Set
 from typing import Type # For some reasons the linter doesn't recognize the use in apply_definition but PyCharm needs it imported to correctly recognize it # pylint: disable=unused-import
 
 from itanium_demangler import parse
@@ -1430,8 +1430,7 @@ class Function(Serializable):
         if not library.has_prototype(self.name):
             return
 
-        proto = library.prototypes[self.name]
-
+        proto = library.get_prototype(self.name)
         self.prototype = proto
         if self.calling_convention is not None:
             self.calling_convention.args = None
@@ -1448,9 +1447,11 @@ class Function(Serializable):
 
     @property
     def demangled_name(self):
-
         if self.name[0:2] == "_Z":
-            ast = parse(self.name)
+            try:
+                ast = parse(self.name)
+            except NotImplementedError:
+                return self.name
             if ast:
                 return ast.__str__()
         return self.name
@@ -1486,6 +1487,18 @@ class Function(Serializable):
         else:
             raise TypeError("calling_convention has to be one of: [SimCC, type(SimCC), None]")
 
+    def functions_called(self) -> Set['Function']:
+        """
+        :return: The set of all functions that can be reached from the function represented by self.
+        """
+        def _called(function_address) -> Set[int]:
+            successors = set(self._function_manager.callgraph.successors(function_address))
+            called_functions = successors.copy()
+            for s in successors:
+                called_functions |= _called(s)
+            return called_functions
+        return { self._function_manager.function(a) for a in _called(self.addr) }
+
     def copy(self):
         func = Function(self._function_manager, self.addr, name=self.name, syscall=self.is_syscall)
         func.transition_graph = networkx.DiGraph(self.transition_graph)
