@@ -9,8 +9,8 @@ from ...engines.light import SpOffset
 from ...keyed_region import KeyedRegion
 from ...code_location import CodeLocation
 from .atoms import Atom, Register, MemoryLocation, Tmp
-from .definition import Definition
-from .undefined import undefined
+from .definition import Definition, Tag
+from .undefined import UNDEFINED
 from .uses import Uses
 from .dataset import DataSet
 
@@ -95,7 +95,7 @@ class LiveDefinitions:
 
         return state
 
-    def kill_definitions(self, atom: Atom, code_loc: CodeLocation, data: Optional[DataSet]=None, dummy=True) -> None:
+    def kill_definitions(self, atom: Atom, code_loc: CodeLocation, data: Optional[DataSet]=None, dummy=True, tags: Set[Tag]=None) -> None:
         """
         Overwrite existing definitions w.r.t 'atom' with a dummy definition instance. A dummy definition will not be
         removed during simplification.
@@ -105,13 +105,13 @@ class LiveDefinitions:
         :return: None
         """
 
-        data = DataSet(undefined, atom.size)
-        self.kill_and_add_definition(atom, code_loc, data, dummy=dummy)
+        data = DataSet(UNDEFINED, atom.size)
+        self.kill_and_add_definition(atom, code_loc, data, dummy=dummy, tags=tags)
 
     def kill_and_add_definition(self, atom: Atom, code_loc: CodeLocation, data: Optional[DataSet],
-                                dummy=False) -> Optional[Definition]:
-        data = data or DataSet(undefined, atom.size)
-        definition: Definition = Definition(atom, code_loc, data, dummy=dummy)
+                                dummy=False, tags: Set[Tag]=None) -> Optional[Definition]:
+        data = data or DataSet(UNDEFINED, atom.size)
+        definition: Definition = Definition(atom, code_loc, data, dummy=dummy, tags=tags)
 
         # set_object() replaces kill (not implemented) and add (add) in one step
         if isinstance(atom, Register):
